/*
 * Definitions of bitree type
 * Internal structure:
 * Number of child of each level is a position
 * bit in corresponding element of array 
 * Teodor Sigaev <teodor@stack.net>
 */

#include "postgres.h"

#include "access/gist.h"
#include "access/itup.h"
#include "access/rtree.h"
#include "utils/elog.h"
#include "utils/palloc.h"
#include "utils/builtins.h"
#include "storage/bufpage.h"
#include <ctype.h>

#include "bitree.h"

PG_FUNCTION_INFO_V1(bitree_in);
PG_FUNCTION_INFO_V1(bitree_out);

Datum	bitree_in(PG_FUNCTION_ARGS);
Datum	bitree_out(PG_FUNCTION_ARGS);

/* compare functions */
PG_FUNCTION_INFO_V1(bitree_cmp);
PG_FUNCTION_INFO_V1(bitree_lt);
PG_FUNCTION_INFO_V1(bitree_le);
PG_FUNCTION_INFO_V1(bitree_eq);
PG_FUNCTION_INFO_V1(bitree_ne);
PG_FUNCTION_INFO_V1(bitree_ge);
PG_FUNCTION_INFO_V1(bitree_gt);
PG_FUNCTION_INFO_V1(bitree_level);
PG_FUNCTION_INFO_V1(bitree_next);
PG_FUNCTION_INFO_V1(bitree_isparent);
PG_FUNCTION_INFO_V1(bitree_risparent);
Datum	bitree_cmp(PG_FUNCTION_ARGS);
Datum	bitree_lt(PG_FUNCTION_ARGS);
Datum	bitree_le(PG_FUNCTION_ARGS);
Datum	bitree_eq(PG_FUNCTION_ARGS);
Datum	bitree_ne(PG_FUNCTION_ARGS);
Datum	bitree_ge(PG_FUNCTION_ARGS);
Datum	bitree_gt(PG_FUNCTION_ARGS);
Datum	bitree_level(PG_FUNCTION_ARGS);
Datum	bitree_next(PG_FUNCTION_ARGS);
Datum	bitree_isparent(PG_FUNCTION_ARGS);
Datum	bitree_risparent(PG_FUNCTION_ARGS);

Datum 
bitree_in(PG_FUNCTION_ARGS) {
	char    *buf = (char *) PG_GETARG_POINTER(0);
	int4	depth=1, palen, len=0, tmp;
	register char *ptr;
	bitreenode *node;
	leveltype *prsnode;
	register int4 state = PRS_WAITNUM, i=0;
	char numbuf[NUMLEN];

	ptr = buf;
	while( ptr && *ptr ) {
		if ( *ptr=='.' )
			depth++;
		ptr++;
	} 

	prsnode = (leveltype*)palloc( sizeof(leveltype) * depth );
	ptr = buf;
	while( 1 && *buf ) {
		if ( state == PRS_WAITNUM ) {
			if ( !isdigit( (unsigned char) *ptr ) ) {
				if ( *ptr ) 
					elog(ERROR, "Unknown character: '%c'", *ptr);
				else
					elog(ERROR, "Unexpected end of string");
			}
			numbuf[0] = *ptr; i=1;
			state = PRS_INNUM;
		} else if ( state == PRS_INNUM ) {
			if ( isdigit( (unsigned char) *ptr ) ) {
				if ( i>=NUMLEN-1 ) 
					elog(ERROR, "Number is too long");
				numbuf[i] = *ptr; i++;
			} else if ( *ptr == '.' || !*ptr ) {
				numbuf[i] = '\0';
				tmp = atoi( numbuf );
				if ( tmp > MAXCHILD || tmp <= 0) 
					elog(ERROR, "Invalid number of child: %d, ( 0<n<=%d )", tmp, MAXCHILD);
				prsnode[len] = 0x1<<(tmp-1);	
				len++;
				if ( !*ptr )
					break;
				state = PRS_WAITNUM;
			} else
				elog(ERROR, "Unknown character: '%c'", *ptr);
		} else
			 elog(ERROR, "Inner logic error");
		ptr++;
	}

	palen = CALCSIZE(len);
	node = (bitreenode*)palloc( palen );
	node->len = palen;
	memcpy( (void*)node->id, (void*)prsnode, sizeof(leveltype)*len ); 
	pfree( prsnode );
	
	PG_RETURN_POINTER(node);
}

Datum
bitree_out(PG_FUNCTION_ARGS) {
	bitreenode *node = (bitreenode*) PG_GETARG_POINTER(0);
	register int4	len = GETNODELEN(node);
	register int4 i,j;
	char *buf;
	register char *ptr;
	
	ptr = buf = (char*)palloc( sizeof(char)*12*len+1 );
	*buf = '\0';
	for( i=0; i<len; i++ ) {
		for(j=0; j< MAXCHILD && (0x01 & (node->id[i] >> j)) == 0; j++);
		
		sprintf(ptr,"%d", j+1);
		ptr = strchr( ptr, '\0' );
		if ( i!=len-1 ) {
			*ptr = '.';
			ptr++;
		}
	}
	buf = (char*)repalloc( (void*)buf, sizeof(char) * ( ptr-buf+1 ) );
	PG_RETURN_POINTER(buf);
}

int 
cmp_binode(bitreenode *n1, bitreenode *n2) {
	int len1 = GETNODELEN(n1);
	int len2 = GETNODELEN(n2);
	register leveltype *ln1,*ln2,*ptrn1, *ptrn2;
	
	ln1 = ptrn1 = n1->id;
	ln2 = ptrn2 = n2->id;

	while( ptrn1-ln1 < len1 && ptrn2-ln2 < len2 ) {
		if ( *ptrn1 == *ptrn2 ) {
			ptrn1++; ptrn2++;
		} else if ( *ptrn1>*ptrn2 )
			return 1;
		else 
			return -1;
	}

	return len1 - len2;	
}
		
Datum
bitree_cmp(PG_FUNCTION_ARGS) {
	PG_RETURN_INT32(
		cmp_binode(
			(bitreenode*)PG_GETARG_POINTER(0),
			(bitreenode*)PG_GETARG_POINTER(1)
		)
	);
}

Datum
bitree_lt(PG_FUNCTION_ARGS) {
	PG_RETURN_BOOL(
		( cmp_binode((bitreenode*)PG_GETARG_POINTER(0), (bitreenode*)PG_GETARG_POINTER(1) ) < 0 ) 
			? true : false
	);
}

Datum
bitree_le(PG_FUNCTION_ARGS) {
	PG_RETURN_BOOL(
		( cmp_binode((bitreenode*)PG_GETARG_POINTER(0), (bitreenode*)PG_GETARG_POINTER(1) ) <= 0 ) 
			? true : false
	);
}

Datum
bitree_eq(PG_FUNCTION_ARGS) {
	PG_RETURN_BOOL(
		( cmp_binode((bitreenode*)PG_GETARG_POINTER(0), (bitreenode*)PG_GETARG_POINTER(1) ) == 0 ) 
			? true : false
	);
}

Datum
bitree_ne(PG_FUNCTION_ARGS) {
	PG_RETURN_BOOL(
		( cmp_binode((bitreenode*)PG_GETARG_POINTER(0), (bitreenode*)PG_GETARG_POINTER(1) ) != 0 ) 
			? true : false
	);
}

Datum
bitree_ge(PG_FUNCTION_ARGS) {
	PG_RETURN_BOOL(
		( cmp_binode((bitreenode*)PG_GETARG_POINTER(0), (bitreenode*)PG_GETARG_POINTER(1) ) >= 0 ) 
			? true : false
	);
}

Datum
bitree_gt(PG_FUNCTION_ARGS) {
	PG_RETURN_BOOL(
		( cmp_binode((bitreenode*)PG_GETARG_POINTER(0), (bitreenode*)PG_GETARG_POINTER(1) ) > 0 ) 
			? true : false
	);
}

Datum
bitree_level(PG_FUNCTION_ARGS) {
	PG_RETURN_INT32( GETNODELEN(PG_GETARG_POINTER(0)) );
}

Datum
bitree_next(PG_FUNCTION_ARGS) {
	bitreenode *newnode, *node = (bitreenode*)PG_GETARG_POINTER(0);
	int4 last = GETNODELEN(node)-1;

	if ( last < 0 )
		elog(ERROR,"Only one root");
	newnode = (bitreenode*)palloc( node->len );
	memcpy( (void*)newnode, (void*)node, node->len );
	newnode->id[ last ] <<= 1;
	if ( newnode->id[ last ] == 0 )
		elog(ERROR,"Max child per node exceed");
	PG_RETURN_POINTER(newnode);
}

bool
inner_bi_isparent(bitreenode *parent, bitreenode *child) {
	int len1 = GETNODELEN(parent);
	int len2 = GETNODELEN(child);
	register leveltype *ln1,*ptrn1, *ptrn2;

	if ( len1 > len2 )
		return false;
	ln1 = ptrn1 = parent->id;
	ptrn2 = child->id;

	while( ptrn1-ln1 < len1 ) {
		if ( *ptrn1 != *ptrn2 )
			return false;
		ptrn1++;
		ptrn2++;
	}

	return true;
}

Datum
bitree_isparent(PG_FUNCTION_ARGS) {
        PG_RETURN_BOOL(
                inner_bi_isparent(
                        (bitreenode*)PG_GETARG_POINTER(0),
                        (bitreenode*)PG_GETARG_POINTER(1)
                )
        );
}

Datum
bitree_risparent(PG_FUNCTION_ARGS) {
        PG_RETURN_BOOL(
                inner_bi_isparent(
                        (bitreenode*)PG_GETARG_POINTER(1),
                        (bitreenode*)PG_GETARG_POINTER(0)
                )
        );
}


