/*
 * Definitions for bitreequery type
 * Internal structure as bitree type, but element of array
 * can have several bits.
 * Teodor Sigaev <teodor@stack.net>
 */

#include "postgres.h"

#include "access/gist.h"
#include "access/itup.h"
#include "access/rtree.h"
#include "utils/elog.h"
#include "utils/palloc.h"
#include "utils/builtins.h"
#include "storage/bufpage.h"
#include <ctype.h>

#include "bitree.h"

PG_FUNCTION_INFO_V1(bitreequery_in);
PG_FUNCTION_INFO_V1(bitreequery_out);

Datum	bitreequery_in(PG_FUNCTION_ARGS);
Datum	bitreequery_out(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(bitq_contains);
PG_FUNCTION_INFO_V1(bitq_contain);

Datum   bitq_contains(PG_FUNCTION_ARGS);
Datum   bitq_contain(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(bitree2query);
Datum   bitree2query(PG_FUNCTION_ARGS);

Datum 
bitreequery_in(PG_FUNCTION_ARGS) {
	char    *buf = (char *) PG_GETARG_POINTER(0);
	int4	depth=1, palen, len=0, tmp;
	register char *ptr;
	bitreenode *node;
	leveltype *prsnode;
	register int4 state = PRS_WAITLEVEL, i=0;
	char numbuf[NUMLEN];
	bool waszero = false;

	ptr = buf;
	while( ptr && *ptr ) {
		if ( *ptr=='.' )
			depth++;
		ptr++;
	} 

	prsnode = (leveltype*)palloc( sizeof(leveltype) * depth );
	MemSet( (void*)prsnode, 0, sizeof(leveltype) * depth );
	ptr = buf;
	while( 1 ) {
		if ( state == PRS_WAITLEVEL ) {
			if ( isdigit( (unsigned char) *ptr ) ) {
				state = PRS_WAITNUM;
				continue;
			} else if ( *ptr=='*' ) {
				prsnode[len] = ~(leveltype)0;
				state = PRS_WAITDELIM;
			} else {
				if ( *ptr )
					elog(ERROR, "Unknown character: '%c'", *ptr);
				else
					elog(ERROR, "Unexpected end of string");
			}
		} else if ( state == PRS_WAITNUM ) {
			if ( !isdigit( (unsigned char) *ptr ) ) {
				if ( *ptr )
					elog(ERROR, "Unknown character: '%c'", *ptr);
				else
					elog(ERROR, "Unexpected end of string");
			}
			numbuf[0] = *ptr; i=1;
			state = PRS_INNUM;
		} else if ( state == PRS_INNUM ) {
			if ( isdigit( (unsigned char) *ptr ) ) {
				if ( i>=NUMLEN-1 )
					elog(ERROR, "Number is too long");
				numbuf[i] = *ptr; i++;
			} else if ( *ptr == '.' || *ptr == '|' || !*ptr ) {
				numbuf[i] = '\0';
				tmp = atoi( numbuf );
				if ( tmp > MAXCHILD || tmp < 0 )
					elog(ERROR, "Invalid number of child: %d, ( 0<=n<=%d )", tmp, MAXCHILD);
				if ( tmp && !waszero )
					prsnode[len] |= 0x1<<(tmp-1);
				else if ( prsnode[len] || waszero )
					elog(ERROR, "Zero can't be used as alternative");
				else
					waszero = true;
				if ( *ptr == '|' ) {
					state = PRS_WAITNUM;
				} else if ( *ptr == '.' || !*ptr) {
					state = PRS_WAITDELIM;
					continue;
				} else
					elog(ERROR, "Inner logic error");
			} else 
				elog(ERROR, "Unknown character: '%c'", *ptr);
		} else if ( state == PRS_WAITDELIM ) {
			if ( *ptr == '.' ) {
				waszero = false;
				len++;
				state = PRS_WAITLEVEL;
			} else if ( !*ptr ) {
				len++;
				break;
			} else
				elog(ERROR, "Unknown character: '%c'", *ptr);
		} else
			elog(ERROR, "Inner logic error");
		ptr++; 
	}

	/* some shorter */
	while( len>0 && prsnode[len-1] == ~(leveltype)0 )
		len--; 
	while( len>1 && prsnode[len-1] == (leveltype)0 && prsnode[len-2] == (leveltype)0 )
		len--;
	/* check */
	if ( len == 1 && prsnode[0] == 0 ) 
		elog(ERROR, "Such bitreequery will produce always void result");
	for( i=0; i<len-1 ; i++ ) 
		if ( prsnode[i]==0 )
			elog(ERROR, "Zero can't be used in middle");
 
	palen = CALCSIZE(len);
	node = (bitreenode*)palloc( palen );
	node->len = palen;
	memcpy( (void*)node->id, (void*)prsnode, sizeof(leveltype)*len ); 
	pfree( prsnode );
	
	PG_RETURN_POINTER(node);
}

Datum
bitreequery_out(PG_FUNCTION_ARGS) {
	bitreenode *node = (bitreenode*) PG_GETARG_POINTER(0);
	register int4	len = GETNODELEN(node);
	register int4 i,j;
	char *buf;
	register char *ptr;
	int first;
	
	ptr = buf = (char*)palloc( sizeof(char)*NUMLEN * MAXCHILD * len +1 );
	*buf = '\0';
	for( i=0; i<len; i++ ) {
		first = 0;
		if ( node->id[i] == ~(leveltype)0 ) {
			*ptr = '*';
			ptr++;
			*ptr = '\0';
		} else if ( node->id[i] == (leveltype)0 ) {
			*ptr = '0';
			ptr++;
			*ptr = '\0';
		} else 
			for(j=0; j< MAXCHILD ; j++) {
				if ( (0x01 & (node->id[i] >> j)) ) {
					if ( first )
						sprintf(ptr,"|%d", j+1);
					else
						sprintf(ptr,"%d", j+1);
					first = 1;
					ptr = strchr( ptr, '\0' );
				}
			}
		
		if ( i!=len-1 ) {
			*ptr = '.';
			ptr++;
		}
	}
	buf = (char*)repalloc( (void*)buf, sizeof(char) * ( ptr-buf+1 ) );
	PG_RETURN_POINTER(buf);
}

char*
printbinode( bitreenode *n ) {
        return (char*)DatumGetPointer(
                DirectFunctionCall1(
                        bitreequery_out,
                        PointerGetDatum(n)
                )
        );
}

bool 
inner_bitq_contain(bitreenode *node, bitreenode *mask, bool ignorezero ) {
	register leveltype *s1 = mask->id, *s2 = node->id;
	register int len1 = GETNODELEN(mask), len2 = GETNODELEN(node);
	register int len = max(len1,len2);

	if ( ignorezero ) {
		do {
			if (  (		
					( (s1 - mask->id < len1 && *s1 ) ? *s1 : ~(leveltype)0 ) 
					& 
					( (s2 - node->id < len2) ? *s2 :  (leveltype)0 ) 
				) == 0 ) {
	
				if ( 
						( (s1 - mask->id < len1) ? *s1 : ~(leveltype)0 ) 
						== 
						( (s2 - node->id < len2) ? *s2 : (leveltype)0 )
					)
						return true;
					else
						return false;	

			} 
	
			s1++; s2++;
			len--;
		} while( len );
	} else {
		do {
			if (  (		
					( (s1 - mask->id < len1) ? *s1 : ~(leveltype)0 ) 
					& 
					( (s2 - node->id < len2) ? *s2 :  (leveltype)0 ) 
				) == 0 ) {
	
				if ( 
						( (s1 - mask->id < len1) ? *s1 : ~(leveltype)0 ) 
						== 
						( (s2 - node->id < len2) ? *s2 : (leveltype)0 )
					)
						return true;
					else
						return false;	

			} 
	
			s1++; s2++;
			len--;
		} while( len );

	}

	return true;	
}

Datum
bitq_contain(PG_FUNCTION_ARGS) {
	PG_RETURN_BOOL(
		inner_bitq_contain(
			(bitreenode*) PG_GETARG_POINTER(0),
			(bitreenode*) PG_GETARG_POINTER(1),
			false
		)
	);
}

Datum
bitq_contains(PG_FUNCTION_ARGS) {
	PG_RETURN_BOOL(
		inner_bitq_contain(
			(bitreenode*) PG_GETARG_POINTER(1),
			(bitreenode*) PG_GETARG_POINTER(0),
			false
		)
	);
}

Datum
bitree2query(PG_FUNCTION_ARGS) {
	bitreenode *query, *node = (bitreenode*)PG_GETARG_POINTER(0);
	query = (bitreenode*) palloc( node->len );
	memcpy( (void*)query, (void*)node, node->len ); 
	PG_RETURN_POINTER(query);
}

