/*
 * GiST interface for array of bitree type
 * Teodor Sigaev <teodor@stack.net>
 */


#include "postgres.h"

#include "access/gist.h"
#include "access/itup.h"
#include "access/rtree.h"
#include "utils/elog.h"
#include "utils/palloc.h"
#include "utils/array.h"
#include "utils/builtins.h"
#include "storage/bufpage.h"
#include "access/nbtree.h"

#include "bitree.h"
#define WISH_F(a,b,c) (double)( -(double)(((a)-(b))*((a)-(b))*((a)-(b)))*(c) )

PG_FUNCTION_INFO_V1( g_bitree_compress );
Datum   g_bitree_compress(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( g_bitree_penalty );
Datum   g_bitree_penalty(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( g_bitree_union );
Datum   g_bitree_union(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( g_bitree_picksplit );
Datum   g_bitree_picksplit(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( g_bitree_consistent );
Datum   g_bitree_consistent(PG_FUNCTION_ARGS);

Datum
g_bitree_compress(PG_FUNCTION_ARGS) {
	GISTENTRY *entry = (GISTENTRY *)PG_GETARG_POINTER(0);

	if ( entry->leafkey ) {
		GISTENTRY  *retval = (GISTENTRY*)palloc(sizeof(GISTENTRY));
		ArrayType *arrnode = DatumGetArrayTypeP(entry->key);
		int4 arrlen = ARRNELEMS(arrnode);
		bitreenode **nodes=NULL;
		int4 	tlen,keylen=0;
		register int4 	i,j;
		bitreenode *key=NULL;

		if ( arrlen ) {
			nodes=(bitreenode **)palloc( sizeof(bitreenode*) * arrlen );
			nodes[0] = ARRPTR(arrnode);
			keylen = GETNODELEN( nodes[0] );
			for(i=1;i<arrlen;i++) {
				nodes[i] = NEXTVAL( nodes[i-1] );
				if ( GETNODELEN( nodes[i] ) > keylen )
					keylen = GETNODELEN( nodes[i] );
			} 
		}

		tlen = CALCSIZE(keylen);
		key = (bitreenode*)palloc(tlen);
		MemSet((void*)key,0,tlen);
		key->len = tlen;
		
		for(i=0;i<arrlen;i++)
			for(j=0;j<GETNODELEN( nodes[i] );j++)
				key->id[ j ] |= nodes[i]->id[ j ];

		if ( PointerGetDatum(arrnode) != entry->key )
			pfree(arrnode);
		if ( nodes ) 
			pfree( nodes );
		gistentryinit(*retval, PointerGetDatum(key),
				entry->rel, entry->page,
				entry->offset, key->len, FALSE);
		PG_RETURN_POINTER(retval);
	} else
		PG_RETURN_POINTER(entry);
}

Datum
g_bitree_penalty(PG_FUNCTION_ARGS) {
	bitreenode *origval = (bitreenode*)DatumGetPointer( ( (GISTENTRY *)PG_GETARG_POINTER(0) )->key );
	bitreenode *newval  = (bitreenode*)DatumGetPointer( ( (GISTENTRY *)PG_GETARG_POINTER(1) )->key );
	float  *penalty = (float *)    PG_GETARG_POINTER(2);
	register int4 i;
	int4 len = GETNODELEN(origval), lenn = GETNODELEN( newval );
	int4 subsize, unionsize=0;
	register leveltype *ptro=origval->id, *ptrn=newval->id;
	leveltype valtmp;

	len = min(len, lenn);
	while( ptrn - newval->id < len ) {
		valtmp = *ptro | *ptrn;
		subsize = 0;	
		for(i=0;i<MAXCHILD;i++)
			subsize += ( ( valtmp>>i ) & 0x01 ) - ( ( *ptro>>i ) & 0x01 );
		unionsize += subsize * ( len - (ptrn - newval->id) );
		ptro++; ptrn++;
	}

	while( ptrn - newval->id < lenn ) {
		subsize=0;
		for(i=0;i<MAXCHILD;i++)
			unionsize += ( *ptrn>>i ) & 0x01;
		unionsize += subsize * ( lenn - (ptrn - newval->id) );
		ptrn++;
	}

	*penalty = (float)unionsize;
	PG_RETURN_POINTER( penalty );
}

Datum
g_bitree_union(PG_FUNCTION_ARGS) {
	bytea *entryvec = (bytea *) PG_GETARG_POINTER(0);
	int *size = (int *) PG_GETARG_POINTER(1);
	int4 lenvec = (VARSIZE(entryvec) - VARHDRSZ) / sizeof(GISTENTRY);
	int4 lenunion=0,lentmp=0;
	register int4 i,j;
	bitreenode *result;
	leveltype *cur;
	
	for(i=0;i<lenvec;i++)
		if ( (lentmp=GETNODELEN(GETENTRY(entryvec,i))) > lenunion )
			lenunion=lentmp;

	*size = CALCSIZE(lenunion);
	result = (bitreenode*)palloc( *size );
	MemSet( (void*)result,0,*size );
	result->len = *size;

	for(i=0;i<lenvec;i++) {
		cur = ( (bitreenode*)GETENTRY(entryvec,i) )->id;
		lentmp = GETNODELEN( GETENTRY(entryvec,i) );
		for(j=0;j<lentmp;j++)
			result->id[j] |= cur[j];
	}

	PG_RETURN_POINTER( result );
}

static int4
getbitsize( bitreenode *node ) {
	register int4 i,subsize=0,len = GETNODELEN(node);
	register leveltype *ptr = node->id;
	int4 size=0;

	while( ptr - node->id < len ) {
		subsize = 0;
		for(i=0;i<MAXCHILD;i++)
			subsize += ( (*ptr)>>i ) & 0x01;
		size += subsize * ( len-(ptr - node->id) );
		ptr++;
	}
	
	return size;
}

static bitreenode *
addtounion( bitreenode *un, bitreenode *n, int4 *size ) {
	register int4 i;
	int4 len = GETNODELEN(n);
	bitreenode *res = (bitreenode*)palloc( un->len );
	register leveltype *ptru = un->id, *ptrn=n->id, *ptrr = res->id;
	int4 subsize;

	res->len = un->len;
	*size=0;
	while( ptrn - n->id < len ) {
		*ptrr = *ptrn | *ptru;
		subsize = 0;
		for(i=0;i<MAXCHILD;i++)
			subsize += ( (*ptrr)>>i ) & 0x01;
		*size += subsize * ( GETNODELEN(un)-(ptrn - n->id) );
		ptrr++; ptrn++; ptru++;
	}
	len = GETNODELEN(un);
	while( ptru - un->id < len ) {
		*ptrr = *ptru;
		subsize = 0;
		for(i=0;i<MAXCHILD;i++)
			subsize += ( (*ptrr)>>i ) & 0x01;
		*size += subsize * ( len-(ptru - un->id) );
		ptrr++; ptru++;
	} 
	return res;
}

typedef struct {
        OffsetNumber    pos;
        float           cost;
} SPLITCOST;

static int
comparecost( const void *a, const void *b ) {
        if ( ((SPLITCOST*)a)->cost == ((SPLITCOST*)b)->cost )
                return 0;
        else
                return ( ((SPLITCOST*)a)->cost > ((SPLITCOST*)b)->cost ) ? 1 : -1;
}

Datum
g_bitree_picksplit(PG_FUNCTION_ARGS) {
	bytea *entryvec = (bytea*) PG_GETARG_POINTER(0);
	GIST_SPLITVEC *v = (GIST_SPLITVEC*) PG_GETARG_POINTER(1);
	register int4 i,j,k;
	register leveltype *ptr_alpha, *ptr_beta;
	bitreenode *datum_l, *datum_r, *datum_alpha, *datum_beta, *union_dr, *union_dl;
	bool            firsttime = true;
	int4    size_alpha,size_beta,sizeu,sizei,subsize;
	int4    size_waste, waste = 0.0;
	int4    size_l, size_r;
	int4    nbytes,maxtlen=0;
	leveltype valtmpu, valtmpi;
	OffsetNumber seed_1=0, seed_2=0;
	OffsetNumber    *left, *right;
	OffsetNumber maxoff;
	SPLITCOST       *costvector;
	
	maxoff = ((VARSIZE(entryvec) - VARHDRSZ) / sizeof(GISTENTRY)) - 2;
	nbytes = (maxoff + 2) * sizeof(OffsetNumber);
	v->spl_left = (OffsetNumber *) palloc(nbytes);
	v->spl_right = (OffsetNumber *) palloc(nbytes);
	
	firsttime = true;
	waste = 0.0;
	maxtlen = GETNODELEN(GETENTRY(entryvec,maxoff));
	/* find to key with largest distance */
	for (i = FirstOffsetNumber; i < maxoff; i = OffsetNumberNext(i)) {
		datum_alpha = GETENTRY(entryvec,i);
		size_alpha = GETNODELEN(datum_alpha);
		if ( size_alpha>maxtlen )
			maxtlen = size_alpha;
		for (j = OffsetNumberNext(i); j <= maxoff; j = OffsetNumberNext(j)) {
			datum_beta = GETENTRY(entryvec,j);
			size_beta = GETNODELEN(datum_beta);
			sizeu =0; sizei=0;
			ptr_alpha = datum_alpha->id;
			ptr_beta = datum_beta->id;
			while( ptr_alpha - datum_alpha->id < min(size_alpha,size_beta) ) {
				valtmpu = *ptr_alpha | *ptr_beta;   
				valtmpi = *ptr_alpha & *ptr_beta;   
				for(k=0;k<MAXCHILD;k++) {
					sizeu += ( valtmpu>>k ) & 0x01;
					sizei += ( valtmpi>>k ) & 0x01;
				}
				ptr_alpha++; ptr_beta++;
			}
			if ( size_alpha > size_beta )
				while( ptr_alpha - datum_alpha->id < size_alpha ) {
					for(k=0;k<MAXCHILD;k++)
						sizeu += ( (*ptr_alpha)>>k ) & 0x01;
					ptr_alpha++;
				}
			else
				while( ptr_beta - datum_beta->id < size_beta ) {
					for(k=0;k<MAXCHILD;k++)
						sizeu += ( (*ptr_beta)>>k ) & 0x01;
					ptr_beta++;
				}
			size_waste = sizeu - sizei;
                        /*
                         * are these a more promising split that what we've already
                         * seen?
                         */
			if (size_waste > waste || firsttime) {
				waste = size_waste;
				seed_1 = i;
				seed_2 = j;
				firsttime = false;
			}
		}
	}

	left = v->spl_left;
	v->spl_nleft = 0;
	right = v->spl_right;
	v->spl_nright = 0;

	nbytes = CALCSIZE(maxtlen);

	datum_alpha = GETENTRY(entryvec,seed_1);
	datum_l = (bitreenode*)palloc( nbytes );
	MemSet( (void*)datum_l, 0, nbytes );
	memcpy( (void*)datum_l, (void*)datum_alpha, datum_alpha->len );
	size_l = getbitsize( datum_l );

	datum_beta = GETENTRY(entryvec,seed_2);
	datum_r = (bitreenode*)palloc( nbytes );
	MemSet( (void*)datum_r, 0, nbytes );
	memcpy( (void*)datum_r, (void*)datum_beta, datum_beta->len );
	size_r = getbitsize( datum_r );

	datum_l->len = datum_r->len = nbytes;

	/* ascending sort entries by cost insertion */
	maxoff = OffsetNumberNext(maxoff);
	costvector=(SPLITCOST*)palloc( sizeof(SPLITCOST)*maxoff );
	for (i = FirstOffsetNumber; i <= maxoff; i = OffsetNumberNext(i)) {
		costvector[i-1].pos = i;
		size_alpha = size_beta = 0;
		datum_alpha = GETENTRY(entryvec,i);
		sizeu = GETNODELEN(datum_alpha);

		ptr_alpha = datum_alpha->id;
		ptr_beta = datum_l->id;
		while( ptr_beta - datum_l->id < sizeu ) {
			valtmpu = *ptr_alpha | *ptr_beta;
			subsize=0;
			for(j=0;j<MAXCHILD;j++)
				subsize += ( valtmpu>>j ) & 0x01;
			size_alpha += subsize * ( maxtlen-(ptr_beta - datum_l->id) );
			ptr_alpha++; ptr_beta++;
		}
		while( ptr_beta - datum_l->id < maxtlen ) {
			subsize=0;
			for(j=0;j<MAXCHILD;j++)
				subsize += ( (*ptr_beta)>>j ) & 0x01;
			size_alpha += subsize * ( maxtlen-(ptr_beta - datum_l->id) );
			ptr_beta++;
		}

		ptr_alpha = datum_alpha->id;
		ptr_beta = datum_r->id;
		while( ptr_beta - datum_r->id < sizeu ) {
			valtmpu = *ptr_alpha | *ptr_beta;
			subsize=0;
			for(j=0;j<MAXCHILD;j++)
				subsize += ( valtmpu>>j ) & 0x01;
			size_beta += subsize * ( maxtlen-(ptr_beta - datum_l->id) );
			ptr_alpha++; ptr_beta++;
		}
		while( ptr_beta - datum_r->id < maxtlen ) {
			subsize=0;
			for(j=0;j<MAXCHILD;j++)
				subsize += ( (*ptr_beta)>>j ) & 0x01;
			size_beta += subsize * ( maxtlen-(ptr_beta - datum_l->id) );
			ptr_beta++;
		}

		costvector[i-1].cost = abs( (size_alpha - size_l) - (size_beta - size_r) );
	}
	qsort( (void*)costvector, maxoff, sizeof(SPLITCOST), comparecost );

	/* insert entries */
	for (j = 0; j < maxoff; j++) {
		i = costvector[j].pos;
		if (i == seed_1) {
			*left++ = i;
			v->spl_nleft++;
			continue;
		} else if (i == seed_2) {
			*right++ = i;
			v->spl_nright++;
			continue;
		}

		datum_alpha = GETENTRY(entryvec,i);
		union_dl = addtounion( datum_l, datum_alpha, &size_alpha);
		union_dr = addtounion( datum_r, datum_alpha, &size_beta);
		if (size_alpha - size_l < size_beta - size_r + WISH_F(v->spl_nleft, v->spl_nright, 1.0)) {
			pfree(datum_l);
			pfree(union_dr);
			datum_l = union_dl;
			size_l = size_alpha;
			*left++ = i;
			v->spl_nleft++;
		} else {
			pfree(datum_r);
			pfree(union_dl);
			datum_r = union_dr;
			size_r = size_beta;
			*right++ = i;
			v->spl_nright++;
		}
	}
	pfree( costvector );
	*right = *left = FirstOffsetNumber;

	/* clear zero level */
	for(j=maxtlen-1 ; j>=0 && datum_l->id[j]==0 ; j--);
	datum_l->len = CALCSIZE(j+1);
	for(j=maxtlen-1 ; j>=0 && datum_r->id[j]==0 ; j--);
	datum_r->len = CALCSIZE(j+1);
	
	v->spl_ldatum = PointerGetDatum(datum_l);
	v->spl_rdatum = PointerGetDatum(datum_r);

	PG_RETURN_POINTER( v );
}

Datum
g_bitree_consistent(PG_FUNCTION_ARGS) {
	bitreenode *query = (bitreenode*)PG_GETARG_POINTER(1);
	bitreenode *key = (bitreenode*)DatumGetPointer( 
		((GISTENTRY*)PG_GETARG_POINTER(0))->key 
	);

	PG_RETURN_BOOL( inner_bitq_contain( key, query, true ) );
}

