/*
 * GiST interface for array of tree type
 * Teodor Sigaev <teodor@stack.net>
 */


#include "postgres.h"

#include "access/gist.h"
#include "access/itup.h"
#include "access/rtree.h"
#include "utils/elog.h"
#include "utils/palloc.h"
#include "utils/array.h"
#include "utils/builtins.h"
#include "storage/bufpage.h"
#include "access/nbtree.h"

#include "entree.h"
#define WISH_F(a,b,c) (double)( -(double)(((a)-(b))*((a)-(b))*((a)-(b)))*(c) )

PG_FUNCTION_INFO_V1( g_entree_compress );
Datum   g_entree_compress(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( g_entree_penalty );
Datum   g_entree_penalty(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( g_entree_union );
Datum   g_entree_union(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( g_entree_picksplit );
Datum   g_entree_picksplit(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( g_entree_consistent );
Datum   g_entree_consistent(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1( g_entreekey_in );
Datum g_entreekey_in(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( g_entreekey_out );
Datum g_entreekey_out(PG_FUNCTION_ARGS);

Datum
g_entreekey_in(PG_FUNCTION_ARGS) {
        elog(ERROR,"Not implemented");
	PG_RETURN_DATUM(0);
}

Datum
g_entreekey_out(PG_FUNCTION_ARGS) {
        elog(ERROR,"Not implemented");
	PG_RETURN_DATUM(0);
}

Datum
g_entree_compress(PG_FUNCTION_ARGS) {
	GISTENTRY *entry = (GISTENTRY *)PG_GETARG_POINTER(0);

	if ( entry->leafkey ) {
		GISTENTRY  *retval = (GISTENTRY*)palloc(sizeof(GISTENTRY));
		ArrayType *arrnode = DatumGetArrayTypeP(entry->key);
		int4 arrlen = ARRNELEMS(arrnode);
		entreenode **nodes=NULL;
		int4 	tlen,keylen=0;
		register int4 	i,j;
		signnode *key=NULL;

		if ( arrlen ) {
			nodes=(entreenode **)palloc( sizeof(entreenode*) * arrlen );
			nodes[0] = ARRPTR(arrnode);
			keylen = GETNODELEN( nodes[0] );
			for(i=1;i<arrlen;i++) {
				nodes[i] = NEXTVAL( nodes[i-1] );
				if ( GETNODELEN( nodes[i] ) > keylen )
					keylen = GETNODELEN( nodes[i] );
			} 
		}

		tlen = SN_CALCSIZE(keylen);
		key = (signnode*)palloc(tlen);
		MemSet((void*)key,0,tlen);
		key->len = tlen;
		
		for(i=0;i<arrlen;i++)
			for(j=0;j<GETNODELEN( nodes[i] );j++)
				key->sign[ j ] |= ( nodes[i]->id[ j ] % MAXSIGNCHILD );

		if ( PointerGetDatum(arrnode) != entry->key )
			pfree(arrnode);
		if ( nodes ) 
			pfree( nodes );
		gistentryinit(*retval, PointerGetDatum(key),
				entry->rel, entry->page,
				entry->offset, key->len, FALSE);
		PG_RETURN_POINTER(retval);
	} else
		PG_RETURN_POINTER(entry);
}

Datum
g_entree_penalty(PG_FUNCTION_ARGS) {
	signnode *origval = (signnode*)DatumGetPointer( ( (GISTENTRY *)PG_GETARG_POINTER(0) )->key );
	signnode *newval  = (signnode*)DatumGetPointer( ( (GISTENTRY *)PG_GETARG_POINTER(1) )->key );
	float  *penalty = (float *)    PG_GETARG_POINTER(2);
	register int4 i;
	int4 len = SN_GETLEN(origval), lenn = SN_GETLEN( newval );
	int4 subsize, unionsize=0;
	register signlevtype *ptro=origval->sign, *ptrn=newval->sign;
	signlevtype valtmp;

len = sizeof(signlevtype)/sizeof(signlevtype);
len = ( (  origval->len  - ( sizeof(int4) )  ) 
	/ 4 );
	len = min(len, lenn);
	while( ptrn - newval->sign < len ) {
		valtmp = *ptro | *ptrn;
		subsize = 0;	
		for(i=0;i<MAXSIGNCHILD;i++)
			subsize += ( ( valtmp>>i ) & 0x01 ) - ( ( *ptro>>i ) & 0x01 );
		unionsize += subsize * ( len - (ptrn - newval->sign) );
		ptro++; ptrn++;
	}

	while( ptrn - newval->sign < lenn ) {
		subsize=0;
		for(i=0;i<MAXSIGNCHILD;i++)
			unionsize += ( *ptrn>>i ) & 0x01;
		unionsize += subsize * ( lenn - (ptrn - newval->sign) );
		ptrn++;
	}

	*penalty = (float)unionsize;
	PG_RETURN_POINTER( penalty );
}

Datum
g_entree_union(PG_FUNCTION_ARGS) {
	bytea *entryvec = (bytea *) PG_GETARG_POINTER(0);
	int *size = (int *) PG_GETARG_POINTER(1);
	int4 lenvec = (VARSIZE(entryvec) - VARHDRSZ) / sizeof(GISTENTRY);
	int4 lenunion=0,lentmp=0;
	register int4 i,j;
	signnode *result;
	signlevtype *cur;
	
	for(i=0;i<lenvec;i++)
		if ( (lentmp=SN_GETLEN(GETENTRY(entryvec,i))) > lenunion )
			lenunion=lentmp;

	*size = SN_CALCSIZE(lenunion);
	result = (signnode*)palloc( *size );
	MemSet( (void*)result,0,*size );
	result->len = *size;

	for(i=0;i<lenvec;i++) {
		cur = ( (signnode*)GETENTRY(entryvec,i) )->sign;
		lentmp = SN_GETLEN( GETENTRY(entryvec,i) );
		for(j=0;j<lentmp;j++)
			result->sign[j] |= cur[j];
	}

	PG_RETURN_POINTER( result );
}

static int4
getbitsize( signnode *node ) {
	register int4 i,subsize=0,len = SN_GETLEN(node);
	register signlevtype *ptr = node->sign;
	int4 size=0;

	while( ptr - node->sign < len ) {
		subsize = 0;
		for(i=0;i<MAXSIGNCHILD;i++)
			subsize += ( (*ptr)>>i ) & 0x01;
		size += subsize * ( len-(ptr - node->sign) );
		ptr++;
	}
	
	return size;
}

static signnode *
addtounion( signnode *un, signnode *n, int4 *size ) {
	register int4 i;
	int4 len = SN_GETLEN(n);
	signnode *res = (signnode*)palloc( un->len );
	register signlevtype *ptru = un->sign, *ptrn=n->sign, *ptrr = res->sign;
	int4 subsize;

	res->len = un->len;
	*size=0;
	while( ptrn - n->sign < len ) {
		*ptrr = *ptrn | *ptru;
		subsize = 0;
		for(i=0;i<MAXSIGNCHILD;i++)
			subsize += ( (*ptrr)>>i ) & 0x01;
		*size += subsize * ( SN_GETLEN(un)-(ptrn - n->sign) );
		ptrr++; ptrn++; ptru++;
	}
	len = SN_GETLEN(un);
	while( ptru - un->sign < len ) {
		*ptrr = *ptru;
		subsize = 0;
		for(i=0;i<MAXSIGNCHILD;i++)
			subsize += ( (*ptrr)>>i ) & 0x01;
		*size += subsize * ( len-(ptru - un->sign) );
		ptrr++; ptru++;
	} 
	return res;
}

typedef struct {
        OffsetNumber    pos;
        float           cost;
} SPLITCOST;

static int
comparecost( const void *a, const void *b ) {
        if ( ((SPLITCOST*)a)->cost == ((SPLITCOST*)b)->cost )
                return 0;
        else
                return ( ((SPLITCOST*)a)->cost > ((SPLITCOST*)b)->cost ) ? 1 : -1;
}

Datum
g_entree_picksplit(PG_FUNCTION_ARGS) {
	bytea *entryvec = (bytea*) PG_GETARG_POINTER(0);
	GIST_SPLITVEC *v = (GIST_SPLITVEC*) PG_GETARG_POINTER(1);
	register int4 i,j,k;
	register signlevtype *ptr_alpha, *ptr_beta;
	signnode *datum_l, *datum_r, *datum_alpha, *datum_beta, *union_dr, *union_dl;
	bool            firsttime = true;
	int4    size_alpha,size_beta,sizeu,sizei,subsize;
	int4    size_waste, waste = 0.0;
	int4    size_l, size_r;
	int4    nbytes,maxtlen=0;
	signlevtype valtmpu, valtmpi;
	OffsetNumber seed_1=0, seed_2=0;
	OffsetNumber    *left, *right;
	OffsetNumber maxoff;
	SPLITCOST       *costvector;
	
	maxoff = ((VARSIZE(entryvec) - VARHDRSZ) / sizeof(GISTENTRY)) - 2;
	nbytes = (maxoff + 2) * sizeof(OffsetNumber);
	v->spl_left = (OffsetNumber *) palloc(nbytes);
	v->spl_right = (OffsetNumber *) palloc(nbytes);
	
	firsttime = true;
	waste = 0.0;
	maxtlen = SN_GETLEN(GETENTRY(entryvec,maxoff));
	/* find to key with largest distance */
	for (i = FirstOffsetNumber; i < maxoff; i = OffsetNumberNext(i)) {
		datum_alpha = (signnode*)GETENTRY(entryvec,i);
		size_alpha = SN_GETLEN(datum_alpha);
		if ( size_alpha>maxtlen )
			maxtlen = size_alpha;
		for (j = OffsetNumberNext(i); j <= maxoff; j = OffsetNumberNext(j)) {
			datum_beta = (signnode*)GETENTRY(entryvec,j);
			size_beta = SN_GETLEN(datum_beta);
			sizeu =0; sizei=0;
			ptr_alpha = datum_alpha->sign;
			ptr_beta = datum_beta->sign;
			while( ptr_alpha - datum_alpha->sign < min(size_alpha,size_beta) ) {
				valtmpu = *ptr_alpha | *ptr_beta;   
				valtmpi = *ptr_alpha & *ptr_beta;   
				for(k=0;k<MAXSIGNCHILD;k++) {
					sizeu += ( valtmpu>>k ) & 0x01;
					sizei += ( valtmpi>>k ) & 0x01;
				}
				ptr_alpha++; ptr_beta++;
			}
			if ( size_alpha > size_beta )
				while( ptr_alpha - datum_alpha->sign < size_alpha ) {
					for(k=0;k<MAXSIGNCHILD;k++)
						sizeu += ( (*ptr_alpha)>>k ) & 0x01;
					ptr_alpha++;
				}
			else
				while( ptr_beta - datum_beta->sign < size_beta ) {
					for(k=0;k<MAXSIGNCHILD;k++)
						sizeu += ( (*ptr_beta)>>k ) & 0x01;
					ptr_beta++;
				}
			size_waste = sizeu - sizei;
                        /*
                         * are these a more promising split that what we've already
                         * seen?
                         */
			if (size_waste > waste || firsttime) {
				waste = size_waste;
				seed_1 = i;
				seed_2 = j;
				firsttime = false;
			}
		}
	}

	left = v->spl_left;
	v->spl_nleft = 0;
	right = v->spl_right;
	v->spl_nright = 0;

	nbytes = SN_CALCSIZE(maxtlen);

	datum_alpha = (signnode*)GETENTRY(entryvec,seed_1);
	datum_l = (signnode*)palloc( nbytes );
	MemSet( (void*)datum_l, 0, nbytes );
	memcpy( (void*)datum_l, (void*)datum_alpha, datum_alpha->len );
	size_l = getbitsize( datum_l );

	datum_beta = (signnode*)GETENTRY(entryvec,seed_2);
	datum_r = (signnode*)palloc( nbytes );
	MemSet( (void*)datum_r, 0, nbytes );
	memcpy( (void*)datum_r, (void*)datum_beta, datum_beta->len );
	size_r = getbitsize( datum_r );

	datum_l->len = datum_r->len = nbytes;

	/* ascending sort entries by cost insertion */
	maxoff = OffsetNumberNext(maxoff);
	costvector=(SPLITCOST*)palloc( sizeof(SPLITCOST)*maxoff );
	for (i = FirstOffsetNumber; i <= maxoff; i = OffsetNumberNext(i)) {
		costvector[i-1].pos = i;
		size_alpha = size_beta = 0;
		datum_alpha = (signnode*)GETENTRY(entryvec,i);
		sizeu = SN_GETLEN(datum_alpha);

		ptr_alpha = datum_alpha->sign;
		ptr_beta = datum_l->sign;
		while( ptr_beta - datum_l->sign < sizeu ) {
			valtmpu = *ptr_alpha | *ptr_beta;
			subsize=0;
			for(j=0;j<MAXSIGNCHILD;j++)
				subsize += ( valtmpu>>j ) & 0x01;
			size_alpha += subsize * ( maxtlen-(ptr_beta - datum_l->sign) );
			ptr_alpha++; ptr_beta++;
		}
		while( ptr_beta - datum_l->sign < maxtlen ) {
			subsize=0;
			for(j=0;j<MAXSIGNCHILD;j++)
				subsize += ( (*ptr_beta)>>j ) & 0x01;
			size_alpha += subsize * ( maxtlen-(ptr_beta - datum_l->sign) );
			ptr_beta++;
		}

		ptr_alpha = datum_alpha->sign;
		ptr_beta = datum_r->sign;
		while( ptr_beta - datum_r->sign < sizeu ) {
			valtmpu = *ptr_alpha | *ptr_beta;
			subsize=0;
			for(j=0;j<MAXSIGNCHILD;j++)
				subsize += ( valtmpu>>j ) & 0x01;
			size_beta += subsize * ( maxtlen-(ptr_beta - datum_l->sign) );
			ptr_alpha++; ptr_beta++;
		}
		while( ptr_beta - datum_r->sign < maxtlen ) {
			subsize=0;
			for(j=0;j<MAXSIGNCHILD;j++)
				subsize += ( (*ptr_beta)>>j ) & 0x01;
			size_beta += subsize * ( maxtlen-(ptr_beta - datum_l->sign) );
			ptr_beta++;
		}

		costvector[i-1].cost = abs( (size_alpha - size_l) - (size_beta - size_r) );
	}
	qsort( (void*)costvector, maxoff, sizeof(SPLITCOST), comparecost );

	/* insert entries */
	for (j = 0; j < maxoff; j++) {
		i = costvector[j].pos;
		if (i == seed_1) {
			*left++ = i;
			v->spl_nleft++;
			continue;
		} else if (i == seed_2) {
			*right++ = i;
			v->spl_nright++;
			continue;
		}

		datum_alpha = (signnode*)GETENTRY(entryvec,i);
		union_dl = addtounion( datum_l, datum_alpha, &size_alpha);
		union_dr = addtounion( datum_r, datum_alpha, &size_beta);
		if (size_alpha - size_l < size_beta - size_r + WISH_F(v->spl_nleft, v->spl_nright, 1.0)) {
			pfree(datum_l);
			pfree(union_dr);
			datum_l = union_dl;
			size_l = size_alpha;
			*left++ = i;
			v->spl_nleft++;
		} else {
			pfree(datum_r);
			pfree(union_dl);
			datum_r = union_dr;
			size_r = size_beta;
			*right++ = i;
			v->spl_nright++;
		}
	}
	pfree( costvector );
	*right = *left = FirstOffsetNumber;

	/* clear zero level */
	for(j=maxtlen-1 ; j>=0 && datum_l->sign[j]==0 ; j--);
	datum_l->len = SN_CALCSIZE(j+1);
	for(j=maxtlen-1 ; j>=0 && datum_r->sign[j]==0 ; j--);
	datum_r->len = SN_CALCSIZE(j+1);
	
	v->spl_ldatum = PointerGetDatum(datum_l);
	v->spl_rdatum = PointerGetDatum(datum_r);

	PG_RETURN_POINTER( v );
}

static bool
gts_eq( signnode *key, entreenode *query )  {
	int klen = SN_GETLEN(key), qlen = GETNODELEN(query);
	register leveltype *qptr = query->id;
	register signlevtype *kptr = key->sign;

	if ( qlen>klen ) 
		return false;

	while( qptr - query->id < qlen ) {
		if ( (*kptr & ( *qptr % MAXSIGNCHILD )) == 0 )
			return false; 
		qptr++;
		kptr++;
	}
	return true;
}

static bool
gts_contain( signnode *key, treequery *query ) {
	int klen = SN_GETLEN(key), qlen = query->size;
	register levelquery *lptr  = query->query;
	register signlevtype *kptr = key->sign;
	register int4 i,index;

	if ( qlen>klen+1 ) /* +1 for zero-ended query */ 
		return false;

	while( kptr - key->sign < min(qlen,klen) ) {
		if ( lptr->size && lptr->var[0] ) {
			index = 1;
			for(i=0;i<lptr->size;i++)
				if ( *kptr & ( lptr->var[i] % MAXSIGNCHILD ) ) {
					index = 0;
					break;
				}
			if ( index )
				return false;	
		}
		lptr = LQ_NEXT(lptr);
		kptr++;
	}
	return true;
}

Datum
g_entree_consistent(PG_FUNCTION_ARGS) {
	char *query = (char*)PG_GETARG_POINTER(1);
	signnode *key = (signnode*)DatumGetPointer( 
		((GISTENTRY*)PG_GETARG_POINTER(0))->key 
	);
	StrategyNumber strategy = (StrategyNumber) PG_GETARG_UINT16(2);
	bool res = false;

	switch( strategy ) {
		case 6:
		case 7:
			res = gts_contain(key, (treequery*)query);
			break;
		case 8:
		case 9:
			res = gts_eq(key, (entreenode*)query);
			break;
		default:
			elog(ERROR,"Uknown strategy number: %d", strategy);
	}

	PG_RETURN_BOOL( res );
}

