/*
 * GiST interface for tree type
 * Teodor Sigaev <teodor@stack.net>
 */


#include "postgres.h"

#include "access/gist.h"
#include "access/itup.h"
#include "access/rtree.h"
#include "utils/elog.h"
#include "utils/palloc.h"
#include "utils/array.h"
#include "utils/builtins.h"
#include "storage/bufpage.h"
#include "access/nbtree.h"

#include "entree.h"

/* used for sorting */
typedef struct rix
{
        int                     index;
        entreenode       *r;
}       RIX;

PG_FUNCTION_INFO_V1( gentree_compress );
Datum   gentree_compress(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( gtree_decompress );
Datum   gtree_decompress(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( gentree_consistent );
Datum   gentree_consistent(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( gentree_union );
Datum   gentree_union(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( gentree_same );
Datum   gentree_same(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( gentree_penalty );
Datum   gentree_penalty(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( gentree_picksplit );
Datum   gentree_picksplit(PG_FUNCTION_ARGS);

typedef struct {
	int4	len;
	int4	isright;
	char	data[1];
} gentreekey;

#define GTK_HDRSIZE	( sizeof(int4) + sizeof(int4) )
#define GET_LKEY(x)     ( (entreenode*)( ((char*)(x)) + GTK_HDRSIZE ) )
#define GET_RCKEY(x)    ( (entreenode*)( ((char*)GET_LKEY(x)) + MAXALIGN(GET_LKEY(x)->len) ) )
#define GET_RKEY(x)     ( ( ((gentreekey*)(x))->isright ) ? GET_RCKEY(x) : GET_LKEY(x) )
#define GTK_CALCSIZE(l,r) 	( GTK_HDRSIZE + MAXALIGN(CALCSIZE(l)) +  ( (r) ? CALCSIZE(r) : 0 ) )

PG_FUNCTION_INFO_V1( gentreekey_in );
Datum gentreekey_in(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1( gentreekey_out );
Datum gentreekey_out(PG_FUNCTION_ARGS);

Datum
gentreekey_in(PG_FUNCTION_ARGS) {
	elog(ERROR,"Not implemented");
	PG_RETURN_DATUM(0);
}

Datum
gentreekey_out(PG_FUNCTION_ARGS) {
	elog(ERROR,"Not implemented");
	PG_RETURN_DATUM(0);
}

Datum
gentree_compress(PG_FUNCTION_ARGS) {
	GISTENTRY  *entry = (GISTENTRY *) PG_GETARG_POINTER(0);

	if (entry->leafkey) {
		GISTENTRY  *retval;
		entreenode *n = (entreenode*)DatumGetPointer( entry->key );
		gentreekey *key;
		int4 len, size = GETNODELEN(n);

		retval = (GISTENTRY*)palloc(sizeof(GISTENTRY));
		len = GTK_CALCSIZE(size,0);
		key = (gentreekey*)palloc( len );
		key->len = len;
		key->isright = false;
		memcpy( (void*)GET_LKEY(key), (void*)n, n->len );
		gistentryinit(*retval, PointerGetDatum(key), entry->rel, entry->page,
			entry->offset, key->len, FALSE);
		PG_RETURN_POINTER(retval);
	}
	PG_RETURN_POINTER(entry);
}

Datum
gtree_decompress(PG_FUNCTION_ARGS) {
	PG_RETURN_DATUM( PG_GETARG_DATUM(0) );
}

static bool
gentq_contain( gentreekey *key, treequery *query ) {
	register leveltype *lptr = GET_LKEY(key)->id, *rptr = GET_RKEY(key)->id;
	int4	llen = GETNODELEN(GET_LKEY(key)), rlen = GETNODELEN(GET_RKEY(key));
	int4	qlen = query->size;
	register levelquery	*qptr = query->query;
	register int4   tlen = min( qlen, min(llen,rlen) );

	if (tlen == 0) {
		if ( qlen == 0 ) 
			return true;
		if ( rlen == 0 )
			return false;
	
		tlen = min(qlen, rlen);	
		do {
			if ( qptr->size ){
				if ( qptr->var[0] ) {
					if ( qptr->var[0] > *rptr )
						return false;
					else 
						return true;
				}
			} else
				return true;
			qptr = LQ_NEXT(qptr);
			rptr++;
			tlen--;
		} while(tlen);
		return true;	
	}
	 
	do {
		if ( qptr->size ){
			if ( qptr->var[0] ) {
				if ( qptr->var[0] > *rptr || qptr->var[ qptr->size-1 ] < *lptr )
					return false;
				else if ( *rptr != *lptr )
					return true;
			}
		} else
			return true;

		qptr = LQ_NEXT(qptr);
		lptr++; rptr++;
		tlen--;
	} while(tlen);
	
	return true;
}

static bool
g_is_parentin( gentreekey *key, entreenode *query ) {
	int4	len = query->len;
	register int4 i;

	for( i=GETNODELEN(query); i>=0 ; i-- ) {
		query->len = CALCSIZE(i);
		if ( cmp_ennode( query, GET_LKEY(key) ) >=0 && cmp_ennode( query, GET_RKEY(key) ) <= 0 ) {
			query->len = len;
			return true;
		}
	}
	query->len = len;

	return false; 
}

Datum   
gentree_consistent(PG_FUNCTION_ARGS) {
	GISTENTRY *entry = (GISTENTRY*)PG_GETARG_POINTER(0);
	entreenode *query = (entreenode*)PG_GETARG_POINTER(1);
	gentreekey *key = (gentreekey*)DatumGetPointer( entry->key );
	StrategyNumber strategy = (StrategyNumber) PG_GETARG_UINT16(2);
	bool res = false;

	switch( strategy ) {
		case BTLessStrategyNumber:
			res = ( GIST_LEAF( entry ) ) ?
					( cmp_ennode(query,GET_LKEY(key)) > 0 )
				:
					( cmp_ennode(query,GET_LKEY(key)) >= 0 );
			break;
		case BTLessEqualStrategyNumber:
			res = ( cmp_ennode( query, GET_LKEY(key) ) >= 0 );  
			break;
		case BTEqualStrategyNumber:
			if ( GIST_LEAF( entry ) ) 
				res = ( cmp_ennode(query,GET_LKEY(key)) == 0 );
			else
				res = (
					cmp_ennode( query, GET_LKEY(key) ) >= 0
					&&
					cmp_ennode( query, GET_RKEY(key) ) <= 0 
				);
			break;
		case BTGreaterEqualStrategyNumber:
			res = ( cmp_ennode( query, GET_RKEY(key) ) <= 0 );  
			break;
		case BTGreaterStrategyNumber:
			res = ( GIST_LEAF( entry ) ) ?
					( cmp_ennode(query,GET_RKEY(key)) < 0 )
				:
					( cmp_ennode(query,GET_RKEY(key)) <= 0 );
			break;
		case 6:
		case 7:
			res = ( GIST_LEAF( entry ) ) ?
				inner_entq_contain( GET_LKEY(key), (treequery*)query )
			:
				gentq_contain(key, (treequery*)query );
			break;
		case 10:
			res = ( GIST_LEAF( entry ) ) ?
				inner_en_isparent( GET_LKEY(key), query )
			:
				g_is_parentin(key, query );
			break;
		case 11:
			do {
				treequery *rq = (treequery*)DatumGetPointer( DirectFunctionCall1(
					entree2query,
					PointerGetDatum(query)
				) ); 
				res = ( GIST_LEAF( entry ) ) ?
					inner_entq_contain( GET_LKEY(key), rq )
				:
					gentq_contain(key, rq );
				pfree(rq);
			} while(0);
			break;
		default:
			elog(ERROR,"Unknown StrategyNumber: %d", strategy);
	}
	PG_RETURN_BOOL(res);
}

Datum
gentree_union(PG_FUNCTION_ARGS) {
	bytea *entryvec = (bytea *) PG_GETARG_POINTER(0);
	int *size = (int *) PG_GETARG_POINTER(1);
	int4 lenvec = (VARSIZE(entryvec) - VARHDRSZ) / sizeof(GISTENTRY);
	register int4 i;
	gentreekey *result;
	register gentreekey *tmp;
	entreenode *min,*max;
	register entreenode *cur;
	
	tmp = (gentreekey*)GETENTRY(entryvec,0);
	min = GET_LKEY(tmp); 
	max = GET_RKEY(tmp); 

	for (i = 1; i < lenvec; i++) {
		tmp = (gentreekey*)GETENTRY(entryvec,i);
		cur = GET_LKEY(tmp);
		if ( cmp_ennode(min,cur) > 0 )
			min = cur;
		cur = GET_RKEY(tmp);
		if ( cmp_ennode(max,cur) < 0 ) 
			max = cur; 
	}

	*size = GTK_HDRSIZE + MAXALIGN(min->len) + max->len;
	result = (gentreekey*)palloc( *size );
	result->len = *size;
	result->isright = true;
	memcpy( (void*)GET_LKEY(result), (void*)min, min->len );
	memcpy( (void*)GET_RKEY(result), (void*)max, max->len );

	PG_RETURN_POINTER( result );
}

Datum
gentree_same(PG_FUNCTION_ARGS) {
	bool *result = (bool *)PG_GETARG_POINTER(2);
	gentreekey *n1 = (gentreekey*)PG_GETARG_POINTER(0);
	gentreekey *n2 = (gentreekey*)PG_GETARG_POINTER(1);

	if ( n1->len != n2->len || n1->isright != n2->isright ) { 
		*result=false;
	} else if ( n1->isright ) {
		*result = ( 
			cmp_ennode(GET_LKEY(n1), GET_LKEY(n2)) == 0
			&&
			cmp_ennode(GET_RKEY(n1), GET_RKEY(n2)) == 0
		) ? true : false;
	} else {
		*result = ( 
			cmp_ennode(GET_LKEY(n1), GET_LKEY(n2)) == 0
		) ? true : false;
	}
	
	PG_RETURN_POINTER(result); 
}

Datum
gentree_penalty(PG_FUNCTION_ARGS) {
	gentreekey *origval = (gentreekey*)DatumGetPointer( ( (GISTENTRY *)PG_GETARG_POINTER(0) )->key );
	gentreekey *newval  = (gentreekey*)DatumGetPointer( ( (GISTENTRY *)PG_GETARG_POINTER(1) )->key );
	float  *penalty = (float *)    PG_GETARG_POINTER(2);
	int4 cmpr,cmpl;

	cmpl = cmp_ennode( GET_LKEY(origval), GET_LKEY(newval) );
	cmpr = cmp_ennode( GET_RKEY(newval), GET_RKEY(origval) );
	*penalty = max( cmpl, 0 ) + max( cmpr, 0 ); 

	PG_RETURN_POINTER(penalty);
}

static int
treekey_cmp(const void *a, const void *b) {
	return cmp_ennode( 
		((RIX *) a)->r,
		((RIX *) b)->r
	);
}

Datum
gentree_picksplit(PG_FUNCTION_ARGS) {
	bytea *entryvec = (bytea*) PG_GETARG_POINTER(0);
	GIST_SPLITVEC *v = (GIST_SPLITVEC*) PG_GETARG_POINTER(1);
	register OffsetNumber i;
	RIX *array;
	OffsetNumber maxoff;
	int	nbytes;
	int    size;
	entreenode *lu_l,*lu_r, *ru_l, *ru_r;
	gentreekey  *lu, *ru;

	maxoff = ((VARSIZE(entryvec) - VARHDRSZ) / sizeof(GISTENTRY)) - 1;
	nbytes = (maxoff + 2) * sizeof(OffsetNumber);
	v->spl_left = (OffsetNumber *) palloc(nbytes);
	v->spl_right = (OffsetNumber *) palloc(nbytes);
	v->spl_nleft = 0; 
	v->spl_nright = 0;
	array = (RIX *) palloc(sizeof(RIX) * (maxoff + 1));

	/* copy the data into RIXes, and sort the RIXes */
	for (i = FirstOffsetNumber; i <= maxoff; i = OffsetNumberNext(i)) {
		array[i].index = i;
		array[i].r = GET_LKEY(GETENTRY( entryvec, i ));
	}

	qsort((void *) &array[FirstOffsetNumber], maxoff - FirstOffsetNumber + 1,
		sizeof(RIX), treekey_cmp);

	lu_l = lu_r = ru_l = ru_r = NULL;
	for (i = FirstOffsetNumber; i <= maxoff; i = OffsetNumberNext(i)) {
		lu = (gentreekey*)GETENTRY( entryvec, array[i].index ); /* use as tmp val */
		if (i <= (maxoff - FirstOffsetNumber + 1) / 2) {
			v->spl_left[v->spl_nleft] = array[i].index;
			v->spl_nleft++;
			if ( lu_r==NULL || cmp_ennode( GET_RKEY(lu), lu_r ) > 0 )
				lu_r = GET_RKEY(lu);
		} else {
			v->spl_right[v->spl_nright] = array[i].index;
			v->spl_nright++;
			if ( ru_r==NULL || cmp_ennode( GET_RKEY(lu), ru_r ) > 0 )
				ru_r = GET_RKEY(lu);
		}
	}

	lu_l = GET_LKEY( GETENTRY( entryvec, array[FirstOffsetNumber].index ) );
	size = GTK_HDRSIZE + MAXALIGN(lu_l->len) + lu_r->len;
	lu = (gentreekey*)palloc( size );
	lu->len = size;
	lu->isright = true;
	memcpy( (void*)GET_LKEY(lu), (void*)lu_l, lu_l->len );
	memcpy( (void*)GET_RKEY(lu), (void*)lu_r, lu_r->len );

	ru_l = GET_LKEY( GETENTRY( entryvec, array[ 1 + ((maxoff - FirstOffsetNumber + 1) / 2) ].index ) );
	size = GTK_HDRSIZE + MAXALIGN(ru_l->len) + ru_r->len;
	ru = (gentreekey*)palloc( size );
	ru->len = size;
	ru->isright = true;
	memcpy( (void*)GET_LKEY(ru), (void*)ru_l, ru_l->len );
	memcpy( (void*)GET_RKEY(ru), (void*)ru_r, ru_r->len );

	pfree(array);
	v->spl_ldatum = PointerGetDatum(lu);
	v->spl_rdatum = PointerGetDatum(ru);

	PG_RETURN_POINTER(v);
}

